import os, textwrap
import numpy as np
import unittest
from context import dpti
# print(dpti.equi)

class TestEquiHeader(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_equi_header_npt(self):
        input = dict(nsteps=1000000, thermo_freq=10, dump_freq=100000, 
            temp=400, tau_t=0.2, 
            tau_p=2.0, mass_map=[118.71], 
            equi_conf='conf.lmp', pres=200000)

        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        PRES            equal 200000.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        """)
        ret2 = dpti.equi.gen_equi_header(**input)
        self.assertEqual(ret1, ret2)

    def test_equi_header_nvt(self):
        input = dict(nsteps=1000000, thermo_freq=10, dump_freq=100000, 
            temp=400, tau_t=0.2, 
            tau_p=2.0, mass_map=[118.71], 
            equi_conf='conf.lmp', pres=None)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        """)
        ret2 = dpti.equi.gen_equi_header(**input)
        self.assertEqual(ret1, ret2)


    def test_equi_header_npt_multi_element(self):
        input = dict(nsteps=1000000, thermo_freq=10, dump_freq=100000, 
            temp=400, tau_t=0.2, 
            tau_p=2.0, mass_map=[118.71, 196.97],
            equi_conf='conf.lmp', pres=None)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        mass            2 196.970000
        """)
        ret2 = dpti.equi.gen_equi_header(**input)
        self.assertEqual(ret1, ret2)

if __name__ == '__main__':
    unittest.main()



import os, textwrap
import numpy as np
import unittest
from unittest.mock import MagicMock, patch, PropertyMock
from context import dpti

class TestEquiEnsembleSetting(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
    
    @patch('numpy.random')
    def test_gen_equi_ensemble_settings_nvt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(ens='nvt')
        ret1 = textwrap.dedent("""\
        fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = dpti.equi.gen_equi_ensemble_settings(**input)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_gen_equi_ensemble_settings_npt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        ret_ensemble_npt = "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n"
        ret_ensemble_npt_aniso = "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}\n"
        ret_ensemble_npt_xy = "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P} couple xy\n"
        ret_ensemble_npt_tri = "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}\n"
        ret_ensemble_nve = "fix             1 all nve\n"
        ret_other = textwrap.dedent("""\
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret_npt = ret_ensemble_npt + ret_other
        ret_npt_aniso = ret_ensemble_npt_aniso + ret_other
        ret_npt_xy = ret_ensemble_npt_xy + ret_other
        ret_npt_tri = ret_ensemble_npt_tri + ret_other
        ret_nve = ret_ensemble_nve + ret_other
        ret2 = dpti.equi.gen_equi_ensemble_settings(ens='npt')
        ret3 = dpti.equi.gen_equi_ensemble_settings(ens='npt-iso')
        ret4 = dpti.equi.gen_equi_ensemble_settings(ens='npt-aniso')
        ret5 = dpti.equi.gen_equi_ensemble_settings(ens='npt-xy')
        ret6 = dpti.equi.gen_equi_ensemble_settings(ens='npt-tri')
        ret7 = dpti.equi.gen_equi_ensemble_settings(ens='nve')
        self.assertEqual(ret_npt, ret2)
        self.assertEqual(ret_npt, ret3)
        self.assertEqual(ret_npt_aniso, ret4)
        self.assertEqual(ret_npt_xy, ret5)
        self.assertEqual(ret_npt_tri, ret6)
        self.assertEqual(ret_nve, ret7)
        with self.assertRaises(RuntimeError):
            dpti.equi.gen_equi_ensemble_settings(ens='foo')


if __name__ == '__main__':
    unittest.main()



import os, textwrap
import numpy as np
import unittest
from context import dpti

class TestEquiThermoSetting(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_gen_equi_dump_settings_false_ave(self):
        input = dict(if_dump_avg_posi=False)
        ret1 = """dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz\n"""
        ret2 = dpti.equi.gen_equi_dump_settings(**input)
        self.assertEqual(ret1, ret2)

    def test_gen_equi_dump_settings_true_ave(self):
        input = dict(if_dump_avg_posi=True)
        ret1 = textwrap.dedent("""\
        compute         ru all property/atom xu yu zu
        fix             ap all ave/atom ${DUMP_FREQ} ${NREPEAT} ${NSTEPS} c_ru[1] c_ru[2] c_ru[3]
        dump            fp all custom ${NSTEPS} dump.avgposi id type f_ap[1] f_ap[2] f_ap[3]
        dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz
        """)
        ret2 = dpti.equi.gen_equi_dump_settings(**input)
        self.assertEqual(ret1, ret2)

if __name__ == '__main__':
    unittest.main()



import os, json, shutil
import numpy as np
import unittest
from context import dpti
from unittest.mock import MagicMock, patch, PropertyMock
from dpti.lib.utils import get_file_md5
from dpti.gdi import _make_tasks_onephase

class TestGdiMakeTask(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.mkdir('tmp_gdi/')

    def setUp(self):
        self.maxDiff = None
        self.test_dir = 'tmp_gdi'
        self.benchmark_dir = 'benchmark_gdi'

    @patch('numpy.random')
    def test_deepmd(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'deepmd/0'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, '../', 'pb.json')
        with open(json_file) as f:
            jdata = json.load(f)

        dpti.gdi._make_tasks_onephase(
            temp=300,
            pres=50000,
            task_path=test_dir,
            jdata=jdata,
            ens='npt',
            conf_file='conf.lmp',
            graph_file='graph.pb',
            if_meam=False,
            meam_model=None
        )

        check_file_list = [ 
            'graph.pb',
            'conf.lmp',
            'in.lammps'
        ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_gdi/')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os
import sys
from enum               import Enum
from subprocess         import Popen, PIPE
from lib.BatchJob import BatchJob
from lib.BatchJob import JobStatus
        
class SlurmJob (BatchJob) :
    def submit_command (self):
        return "sbatch"
    def check_status (self):
        job_id = self.get_job_id ()
        if len(job_id) == 0 :
            return JobStatus.unsubmitted
        ret = Popen (["squeue --job " + job_id], shell=True, stdout=PIPE, stderr=PIPE)
        stdout, stderr = ret.communicate()
        if (ret.returncode != 0) :
            if str("Invalid job id specified") in str(stderr, encoding='ascii') :
                if os.path.exists (self.job_finish_tag) :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                Logger.error ("status command " + "squeue" + " fails to execute")
                Logger.error ("erro info: " + str(stderr, encoding='ascii'))
                Logger.error ("return code: " + str(ret.returncode))
                sys.exit ()
        status_line = str(stdout, encoding='ascii').split ('\n')[-2]
        status_word = status_line.split ()[4]
#        status_word = status_line.split ()[-4]
#        print ("status line: " + status_line)
#        print ("status word: " + status_word)
#        print (status_word)
        if      status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif    status_word in ["R","CG"] :
            return JobStatus.running
        elif    status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            if os.path.exists (self.job_finish_tag) :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown        

if __name__ == "__main__" :
    job = SlurmJob ("/home/han.wang/data/test/string/test", "cu01.sleep")
    job.submit ()
    print ("submit done")
    stat = job.check_status ()
    print ("check done")
    print (stat)
    



#!/usr/bin/env ovitos
'''
This Script is adapted from Alexander Stukowski, the author of OVITO.
See: http://forum.ovito.org/index.php?topic=131.0 for details.
'''
import os
import sys
import argparse
import numpy as np

from ovito.io import *

supp_ofmt = ['lammps_dump', 'lammps_data', 'vasp']
supp_exts = ['dump', 'lmp', 'poscar/POSCAR']

parser = argparse.ArgumentParser()
parser.add_argument("-m", "--ofmt", type=str, 
                    help="the output format, supported: " + str(supp_ofmt))
parser.add_argument("INPUT", type=str, 
                    help="the input file")
parser.add_argument("OUTPUT", type=str, 
                    help="the output file, supported ext: " + str(supp_exts))
args = parser.parse_args()

fin = args.INPUT
fout = args.OUTPUT
if args.ofmt is not None :
    ofmt = args.ofmt
else :
    ext = fout.split('.')[-1]
    if ext == 'dump' :
        ofmt = 'lammps_dump'
    elif ext == 'lmp' :
        ofmt = 'lammps_data'
    elif ext == 'poscar' or ext == 'POSCAR' :
        ofmt = 'vasp'
if not ofmt in supp_ofmt :
    raise RuntimeError ("output format " + ofmt + " is not supported. use one of " + str(supp_ofmt))

columns = None
if ofmt == "lammps_dump" :
    columns=["Particle Identifier", "Particle Type", "Position.X", "Position.Y", "Position.Z"]

node = import_file(fin)
if columns is not None :
    export_file(node, fout, ofmt, columns = columns)
else :
    export_file(node, fout, ofmt)
    



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
import numpy as np
import scipy.constants as pc

from dpti import einstein
from dpti import  hti
from dpti.lib import lmp

def _main ():
    parser = argparse.ArgumentParser(
        description="Compute free energy by Hamiltonian TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')
    parser_gen.add_argument('-s','--switch', type=str, default = 'one-step',
                            choices = ['one-step', 'two-step', 'three-step'],
                            help='one-step: switching on DP and switching off spring simultanenously.\
                            two-step: 1 switching on DP, 2 switching off spring.\
                            three-step: 1 switching on soft LJ, 2 switching on DP, 3 switching off spring and soft LJ.')

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-t','--type', type=str, default = 'helmholtz', 
                             choices=['helmholtz', 'gibbs'], 
                             help='the type of free energy')
    parser_comp.add_argument('-m','--inte-method', type=str, default = 'inte', 
                             choices=['inte', 'mbar'], 
                             help='the method of thermodynamic integration')
    parser_comp.add_argument('-d','--disorder-corr', action = 'store_true',
                             help='apply disorder correction for ice')
    parser_comp.add_argument('-p','--partial-disorder', type=str,
                             choices = ['3', '5'],
                             help='apply partial disorder correction for ice')
    parser_comp.add_argument('-s','--scheme', type=str, default = 'simpson', 
                             help='the numeric integration scheme')
    parser_comp.add_argument('-S','--shift', type=float, default = 0.0, 
                             help='a constant shift in the energy/mole computation, will be removed from FE')
    parser_comp.add_argument('-g', '--pv', type=float, default = None,
                             help='press*vol value override to calculate Gibbs free energy')
    parser_comp.add_argument('-G', '--pv-err', type=float, default = None,
                             help='press*vol error')

    parser_comp = subparsers.add_parser('refine', help= 'Refine the grid of a job')
    parser_comp.add_argument('-i', '--input', type=str, required=True,
                             help='input job')
    parser_comp.add_argument('-o', '--output', type=str, required=True,
                             help='output job')
    parser_comp.add_argument('-e', '--error', type=float, required=True,
                             help='the error required')
    parser_comp.add_argument('-p', '--print', action = 'store_true',
                             help='print the refinement and exit')
    args = parser.parse_args()

    return exec_args(args=args, parser=parser)

def exec_args(args, parser):
    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        with open(args.PARAM, 'r') as j:
            jdata = json.load(j)
        if 'crystal' in jdata and jdata['crystal'] == 'frenkel' :
            print('# gen task with Frenkel\'s Einstein crystal')
        else :
            print('# gen task with Vega\'s Einstein molecule')
        hti.make_tasks(output, jdata, 'einstein', args.switch)
    elif args.command == 'refine' :
        hti.refine_task(args.input, args.output, args.error, args.print)        
    elif args.command == 'compute' :
        job = args.JOB
        jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))
        fp_conf = open(os.path.join(args.JOB, 'conf.lmp'))
        sys_data = lmp.to_system_data(fp_conf.read().split('\n'))
        natoms = sum(sys_data['atom_numbs'])
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
        nmols = natoms // 3
        # compute e0
        if 'crystal' not in jdata :
            jdata['crystal'] = 'vega'
        crystal = jdata['crystal']
        if crystal == 'vega' :
            e0 = einstein.free_energy(job) * 3
        else :
            e0 = einstein.frenkel(job) * 3
        # compute Paulin estimate for disordered entropy
        if args.disorder_corr :
            temp = jdata['temp']
            if args.partial_disorder is not None:
                if args.partial_disorder == '5':
                    pauling_corr = -pc.Boltzmann * temp / pc.electron_volt * 0.3817
                    note_pauling = '(ice5)'
                elif args.partial_disorder == '3':
                    pauling_corr = -pc.Boltzmann * temp / pc.electron_volt * 0.3686
                    note_pauling = '(ice3)'
                else:
                    raise RuntimeError(f'unknow partial_disorder {args.partial_disorder}')
            else:
                pauling_corr = -pc.Boltzmann * temp / pc.electron_volt * np.log(1.5)
                note_pauling = '      '
            e0 += pauling_corr
        else :
            note_pauling = '      '
            pauling_corr = 0
        # compute integration
        de, de_err, thermo_info = hti.post_tasks(job, jdata, natoms = nmols, method = args.inte_method, scheme = args.scheme)
        info = thermo_info.copy()
        # printing
        print_format = '%20.12f  %10.3e  %10.3e'
        hti.print_thermo_info(thermo_info)
        if crystal == 'vega' :
            print('# free ener of Einstein Mole: %20.8f' % (e0))
        else :
            print('# free ener of Einstein Crys: %20.8f' % (e0))
        print('# Pauling corr %s:        %20.8f' % (note_pauling, pauling_corr))
        print(('# fe integration              ' + print_format) \
              % (de, de_err[0], de_err[1]))        
        print( '# fe const shift              %20.12f' % args.shift)
        # if args.type == 'helmholtz' :
        print('# Helmholtz free ener per mol (stat_err inte_err) [eV]:')
        print(print_format % (e0 + de - args.shift, de_err[0], de_err[1]))
        if args.type == 'gibbs' :
            if args.pv is not None:
                pv = args.pv
                print(f"# use manual pv=={pv}")
            else:
                pv = thermo_info['pv']
            if args.pv_err is not None:
                pv_err = args.pv_err
                print(f"# use manual pv_err=={pv_err}")
            else:
                pv_err = thermo_info['pv_err']
            e1 = e0 + de + pv - args.shift
            e1_err = np.sqrt(de_err[0]**2 + pv_err**2)
            print('# Gibbs free ener per mol (stat_err inte_err) [eV]:')
            print(print_format % (e1, e1_err, de_err[1]))
        free_energy_type=args.type
        info['free_energy_type'] = free_energy_type
        info['pv'] = pv
        info['pv_err'] = pv_err
        # info['de'] = de
        # info['de_err'] = de_err
        info['e1'] = e1
        info['e1_err'] = e1_err
        with open(os.path.join(job, 'result.json'), 'w') as result:
            result.write(json.dumps(info))
        return info


if __name__ == '__main__' :
    _main()



import os, textwrap
import numpy as np
import unittest
from unittest.mock import MagicMock, patch, PropertyMock
from context import dpti
from potential_common import meam_model

class TestEquiGenLammpsInput(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    @patch('numpy.random')
    def test_deepmd_npt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(nsteps=1000000, 
            thermo_freq=10, 
            dump_freq=100000, 
            mass_map=[118.71],
            temp=400, 
            tau_t=0.2,
            tau_p=2.0, 
            equi_conf='conf.lmp',
            model='graph.pb',
            timestep=0.002, 
            if_dump_avg_posi=False,
            ens='npt-iso',
            pres=200000,
            if_meam=False,
            meam_model=None,
        )

        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        PRES            equal 200000.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        pair_style      deepmd graph.pb
        pair_coeff
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002000
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]
        dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)

        ret2 = dpti.equi.gen_equi_lammps_input(**input)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_deepmd_nvt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(nsteps=1000000, 
            thermo_freq=10, 
            dump_freq=100000, 
            mass_map=[118.71],
            temp=400, 
            tau_t=0.2,
            tau_p=2.0, 
            equi_conf='conf.lmp',
            model='graph.pb',
            timestep=0.002, 
            if_dump_avg_posi=False,
            ens='nvt',
            pres=200000,
            if_meam=False,
            meam_model=None,
        )
        ret2 = dpti.equi.gen_equi_lammps_input(**input)

        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        PRES            equal 200000.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        pair_style      deepmd graph.pb
        pair_coeff
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002000
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]
        dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz
        fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_meam_npt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(nsteps=1000000, 
            thermo_freq=10, 
            dump_freq=100000, 
            mass_map=[118.71],
            temp=400, 
            tau_t=0.2,
            tau_p=2.0, 
            equi_conf='conf.lmp',
            model=None,
            timestep=0.002, 
            if_dump_avg_posi=False,
            ens='npt-xy',
            pres=200000,
            if_meam=True,
            meam_model=meam_model,
        )
        ret2 = dpti.equi.gen_equi_lammps_input(**input)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 1000000
        variable        THERMO_FREQ     equal 10
        variable        DUMP_FREQ       equal 100000
        variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}
        variable        TEMP            equal 400.000000
        variable        PRES            equal 200000.000000
        variable        TAU_T           equal 0.200000
        variable        TAU_P           equal 2.000000
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        pair_style      meam
        pair_coeff      * * library_18Metals.meam Sn Sn_18Metals.meam Sn
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002000
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]
        dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P} couple xy
        fix             mzero all momentum 10 linear 1 1 1
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        velocity        all zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        self.assertEqual(ret1, ret2)


if __name__ == '__main__':

    unittest.main()



#%%
import os, textwrap, json, shutil
import numpy as np
import unittest
from context import dpti
# from potential_common import soft_param, soft_param_three_element, meam_model
# from dpti.lib.lammps import get_natoms, get_thermo, get_last_dump
# from dpti.lib.dump import from_system_data
from numpy.testing import assert_almost_equal
from unittest.mock import MagicMock, patch, PropertyMock

from dpti import ti_water
from dpti.lib.utils import get_file_md5

class TestTiWaterGenLammpsInput(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    @patch('numpy.random')
    def test_ti_water_gen_tasks(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        args = MagicMock(
            output='tmp_ti_water/new_job/',  
            command='gen',
            PARAM='benchmark_ti_water/ti_water.json'
        )
        ti_water.exec_args(args=args, parser=None)
        check_file_list = [
            'conf.lmp', 
            'task.000003/in.lammps',
            'task.000003/graph.pb',
            'task.000003/thermo.out'
        ]
        for file in check_file_list:
            f1 = os.path.join('benchmark_ti_water/new_job/', file)
            f2 = os.path.join('tmp_ti_water/new_job/', file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_ti_water_old_json_gen_tasks(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        args = MagicMock(
            output='tmp_ti_water/old_json_job/',  
            command='gen',
            PARAM='benchmark_ti_water/ti_water.json.old'
        )
        ti_water.exec_args(args=args, parser=None)
        check_file_list = [
            'conf.lmp', 
            'task.000003/in.lammps',
            'task.000003/graph.pb',
            'task.000003/thermo.out'
        ]
        for file in check_file_list:
            f1 = os.path.join('benchmark_ti_water/new_job/', file)
            f2 = os.path.join('tmp_ti_water/old_json_job/', file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_ti_water/')



import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import meam_model

class TestEquiForceField(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_deepmd(self):
        input = dict(model="graph.pb", if_meam=False, meam_model=None)

        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        pair_style      deepmd graph.pb
        pair_coeff
        """)
        ret2 = dpti.equi.gen_equi_force_field(**input)
        self.assertEqual(ret1, ret2)

    def test_meam(self):
        input = dict(model=None,
            if_meam=True,
            meam_model=meam_model
        )
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        pair_style      meam
        pair_coeff      * * library_18Metals.meam Sn Sn_18Metals.meam Sn
        """)
        ret2 = dpti.equi.gen_equi_force_field(**input)
        self.assertEqual(ret1, ret2)

    def test_meam_raise_error(self):
        input = dict(model=None,
            if_meam=True,
            meam_model={})
        with self.assertRaises(KeyError):
            dpti.equi.gen_equi_force_field(**input)

    def test_meam_raise_error_lack_key(self):
        err_meam_model1 = dict(
            meam_library='library_18Metals.meam', 
            meam_potential='Sn_18Metals.meam', 
        )
        input = dict(model=None,
            if_meam=True,
            meam_model=err_meam_model1)
        with self.assertRaises(KeyError):
            dpti.equi.gen_equi_force_field(**input)

if __name__ == '__main__':
    unittest.main()



import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import soft_param, soft_param_three_element
from dpti.hti_liq import _ff_soft_on

class TestFfSpring(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_one_element(self):
        input = dict(lamb=0.075, sparam=soft_param)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        pair_style      lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      1 1 ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
        """)
        ret2 = _ff_soft_on(**input)
        self.assertEqual(ret1, ret2)

    def test_three_element(self):
        input = dict(lamb=0.075, sparam=soft_param_three_element)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        pair_style      lj/cut/soft 1.000000 0.600000 6.000000
        pair_coeff      1 1 ${EPSILON} 2.000000 0.500000
        pair_coeff      1 2 ${EPSILON} 2.010000 0.500000
        pair_coeff      1 3 ${EPSILON} 2.020000 0.500000
        pair_coeff      2 2 ${EPSILON} 2.110000 0.500000
        pair_coeff      2 3 ${EPSILON} 2.120000 0.500000
        pair_coeff      3 3 ${EPSILON} 2.220000 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
        """)
        ret2 = _ff_soft_on(**input)
        self.assertEqual(ret1, ret2)
        # print(ret2)
        # pass
        # input = dict(lamb=0.075, m_spring_k=[118.71],
        #     var_spring=False)
        # ret1 = textwrap.dedent("""\
        # group           type_1 type 1
        # fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        # fix_modify      l_spring_1 energy yes
        # variable        l_spring equal f_l_spring_1
        # """)
        # ret2 = dpti.hti._ff_spring(**input)
        # self.assertEqual(ret1, ret2)


# s.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# import dpti
# from . import lib
# import .lib
# from . import equi
# from . import hti
# from . import hti_liq
# from . import ti
# from . import gdi
# from dpti import lib



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
import numpy as np
import scipy.constants as pc
import pymbar

# sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../'))
from dpti.lib.utils import create_path, relative_link_file
from dpti.lib.utils import get_first_matched_key_from_dict
from dpti.lib.utils import copy_file_list
from dpti.lib.utils import block_avg
from dpti.lib.utils import integrate_range
# from lib.utils import integrate_sys_err
from dpti.lib.utils import compute_nrefine
from dpti.lib.utils import parse_seq, link_file_in_dict
from dpti.lib.utils import get_task_file_abspath
from dpti.lib.lammps import get_thermo
from dpti.lib.lammps import get_natoms

# from dpti.equi import gen_equi_lammps_input

def make_iter_name (iter_index) :
    return "task_ti." + ('%04d' % iter_index)

def parse_seq_ginv (seq) :
    tmp_seq = parse_seq(seq)
    t_begin = tmp_seq[0]
    t_end = tmp_seq[-1]
    ngrid = len(tmp_seq) - 1
    hh = (1/t_end - 1/t_begin) / ngrid
    inv_grid = np.arange(1/t_begin, 1/t_end+0.5*hh, hh)
    inv_grid = 1./inv_grid
    return inv_grid

def _gen_lammps_input (conf_file, 
                       mass_map,
                       model,
                       nsteps,
                       timestep,
                       ens,
                       temp,
                       pres=1.0, 
                       tau_t=0.1,
                       tau_p=0.5,
                       thermo_freq=100,
                       copies=None,
                       if_meam=False,
                       meam_model=None):
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        NSTEPS          equal %d\n' % nsteps
    ret += 'variable        THERMO_FREQ     equal %d\n' % thermo_freq
    # ret += 'variable        DUMP_FREQ       equal %d\n' % thermo_freq
    ret += 'variable        TEMP            equal %f\n' % temp
    ret += 'variable        PRES            equal %f\n' % pres
    ret += 'variable        TAU_T           equal %f\n' % tau_t
    ret += 'variable        TAU_P           equal %f\n' % tau_p
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      atomic\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % conf_file
    if copies is not None :
        ret += 'replicate       %d %d %d\n' % (copies[0], copies[1], copies[2])
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)) :
        ret+= "mass            %d %f\n" %(jj+1, mass_map[jj])
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    # if if_meam:
    #     ret += 'pair_style      meam \n'
    #     ret += 'pair_coeff      * * /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn\n'
    if if_meam:
        ret += 'pair_style      meam\n'
        ret += f'pair_coeff      * * {meam_model["library"]} {meam_model["element"]} {meam_model["potential"]} {meam_model["element"]}\n'
    else:
        ret += 'pair_style      deepmd %s\n' % model
        ret += 'pair_coeff\n'
    ret += '# --------------------- MD SETTINGS ----------------------\n'    
    ret += 'neighbor        1.0 bin\n'
    ret += 'timestep        %s\n' % timestep
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'compute         allmsd all msd\n'
    if ens == 'nvt' :        
        ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol c_allmsd[*]\n'
    elif 'npt' in ens :
        ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol c_allmsd[*]\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)                
    ret += '# dump            1 all custom ${DUMP_FREQ} traj.dump id type x y z\n'
    if ens == 'nvt' :
        ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
    elif ens == 'npt-iso' or ens == 'npt':
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'npt-aniso' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'npt-tri' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'npt-xy' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P} couple xy\n'
    elif ens == 'nve' :
        ret += 'fix             1 all nve\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)        
    ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
    ret += '# --------------------- INITIALIZE -----------------------\n'    
    ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
    ret += 'velocity        all zero linear\n'
    ret += '# --------------------- RUN ------------------------------\n'    
    ret += 'run             ${NSTEPS}\n'
    ret += 'write_data      out.lmp\n'
    
    return ret


def make_tasks(iter_name, jdata, if_meam=None):
    ti_settings = jdata.copy()
    if if_meam is None:
        if_meam = jdata.get('if_meam', None)
    equi_conf = jdata['equi_conf']
    equi_conf = os.path.abspath(equi_conf)
    copies = None
    if 'copies' in jdata :
        copies = jdata['copies']
    model = jdata['model']
    meam_model = jdata.get('meam_model', None)
    # model = os.path.abspath(model)
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['model_mass_map', 'mass_map'])
    nsteps = jdata['nsteps']
    # timestep = jdata['timestep']
    timestep = get_first_matched_key_from_dict(jdata, ['timestep', 'dt'])
    # thermo_freq = jdata['thermo_freq']
    thermo_freq = get_first_matched_key_from_dict(jdata, ['thermo_freq', 'stat_freq'])
    # thermos = jdata['thermos']
    ens = jdata['ens']
    path = jdata['path']
    if 'nvt' in ens :
        if path == 't' :
            temp_seq = get_first_matched_key_from_dict(jdata, ['temp_seq', 'temps'])
            temp_list = parse_seq(temp_seq)
            tau_t = jdata['tau_t']
            ntasks = len(temp_list)
        else :
            raise RuntimeError('supported path of nvt ens is \'t\'')
    elif 'npt' in ens :
        if path == 't' :
            temp_seq = get_first_matched_key_from_dict(jdata, ['temp_seq', 'temps'])
            temp_list = parse_seq(temp_seq)
            pres = get_first_matched_key_from_dict(jdata, ['pres', 'press'])
            ntasks = len(temp_list)
        elif path == 't-ginv' :
            temp_seq = get_first_matched_key_from_dict(jdata, ['temp_seq', 'temps'])
            temp_list = parse_seq_ginv(temp_seq)
            pres = get_first_matched_key_from_dict(jdata, ['pres', 'press'])
            ntasks = len(temp_list)
        elif path == 'p' :
            temp = get_first_matched_key_from_dict(jdata, ['temp', 'temps'])
            pres_seq = get_first_matched_key_from_dict(jdata, ['pres_seq', 'press'])
            pres_list = parse_seq(pres_seq)
            ntasks = len(pres_list)
        else :
            raise RuntimeError('supported path of npt ens are \'t\' or \'p\'')
        tau_t = jdata['tau_t']
        tau_p = jdata['tau_p']
    else :
        raise RuntimeError('invalid ens')

    job_abs_dir = create_path(iter_name)




    # dct1 = link_file_in_dict(
    #     dct=jdata,
    #     key_list=["equi_conf", "model"],
    #     target_dir=job_abs_dir
    # )
    # ti_settings.update(dct1)

    # meam_model = jdata.get('meam_model', None)
    # dct2 = link_file_in_dict(
    #     dct=meam_model,
    #     key_list=["library", "potential"],
    #     target_dir=job_abs_dir
    # )
    # if meam_model:
    #     ti_settings['meam_model'].update(dct2)

    # link_file_dict = {}
    # link_file_dict.update(dct1)
    # link_file_dict.update(dct2)


    # copied_conf = os.path.join(os.path.abspath(iter_name), 'conf.lmp')
    # shutil.copyfile(equi_conf, copied_conf)
    # jdata['equi_conf'] = 'conf.lmp'
    # linked_model = os.path.join(os.path.abspath(iter_name), 'graph.pb')
    # shutil.copyfile(model, linked_model)
    # jdata['model'] = 'graph.pb'

    # cwd = os.getcwd()
    # os.chdir(iter_name)
    relative_link_file(equi_conf, job_abs_dir)
    if model:
        relative_link_file(model, job_abs_dir)
    if if_meam:
        relative_link_file(meam_model['library'], job_abs_dir)
        relative_link_file(meam_model['potential'], job_abs_dir)

    with open(os.path.join(job_abs_dir, 
        'ti_settings.json'), 'w') as fp:
        json.dump(ti_settings, fp, indent=4)

    with open(os.path.join(job_abs_dir, 'ti_settings.json'),'w') as f:
            json.dump(ti_settings, f, indent=4)

    for ii in range(ntasks) :
        task_dir = os.path.join(job_abs_dir, 'task.%06d' % ii)
        task_abs_dir = create_path(task_dir)
        # os.chdir(work_path)

        relative_link_file(equi_conf, task_abs_dir)
        if model:
            relative_link_file(model, task_abs_dir)
        if if_meam:
            relative_link_file(meam_model['library'], task_abs_dir)
            relative_link_file(meam_model['potential'], task_abs_dir)

        # for file in list(link_file_dict.values()):
        #     file_path = os.path.join(job_abs_dir, file)
        #     relative_link_file(file_path, task_abs_dir)
            # os.symlink()
        # os.symlink(os.path.relpath(copied_conf), 'conf.lmp')
        # os.symlink(os.path.relpath(linked_model), 'graph.pb')
        if 'nvt' in ens and path == 't' :
            lmp_str \
                = _gen_lammps_input(os.path.basename(equi_conf),
                                    mass_map, 
                                    model,
                                    nsteps, 
                                    timestep,
                                    ens,
                                    temp_list[ii],
                                    pres=pres,
                                    tau_t = tau_t,
                                    thermo_freq = thermo_freq, 
                                    copies = copies,
                                    if_meam=if_meam,
                                    meam_model=meam_model
                                    )
            thermo_out = temp_list[ii]
            # with open('thermo.out', 'w') as fp :
            #     fp.write('%f' % temps[ii])
        elif 'npt' in ens and (path == 't' or path == 't-ginv'):
            lmp_str \
                = _gen_lammps_input(os.path.basename(equi_conf),
                                    mass_map, 
                                    model,
                                    nsteps, 
                                    timestep,
                                    ens,
                                    temp_list[ii],
                                    pres,
                                    tau_t = tau_t,
                                    tau_p = tau_p,
                                    thermo_freq = thermo_freq, 
                                    copies = copies,
                                    if_meam=if_meam,
                                    meam_model=meam_model
                                    )
            thermo_out = temp_list[ii]
            # with open('thermo.out', 'w') as fp :
            #     fp.write('%f' % (temps[ii]))
        elif 'npt' in ens and path == 'p' :
            lmp_str \
                = _gen_lammps_input(os.path.basename(equi_conf),
                                    mass_map, 
                                    model,
                                    nsteps, 
                                    timestep,
                                    ens,
                                    temp,
                                    pres_list[ii],
                                    tau_t = tau_t,
                                    tau_p = tau_p,
                                    thermo_freq = thermo_freq, 
                                    copies = copies,
                                    if_meam=if_meam,
                                    meam_model=meam_model
                                )
            thermo_out = pres_list[ii]
        else:
            raise RuntimeError('invalid ens or path setting' )
        
        with open(os.path.join(task_abs_dir, 'thermo.out'), 'w') as fp:
            fp.write('%f' % (thermo_out))
        with open(os.path.join(task_abs_dir, 'in.lammps'), 'w') as fp:
            fp.write(lmp_str)
        
        # os.chdir(cwd)

def _compute_thermo (lmplog, natoms, stat_skip, stat_bsize) :
    data = get_thermo(lmplog)
    ea, ee = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
    ha, he = block_avg(data[:, 4], skip = stat_skip, block_size = stat_bsize)
    ta, te = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
    pa, pe = block_avg(data[:, 6], skip = stat_skip, block_size = stat_bsize)
    va, ve = block_avg(data[:, 7], skip = stat_skip, block_size = stat_bsize)
    thermo_info = {}
    thermo_info['p'] = pa
    thermo_info['p_err'] = pe
    thermo_info['v'] = va / natoms
    thermo_info['v_err'] = ve / np.sqrt(natoms)
    thermo_info['e'] = ea / natoms
    thermo_info['e_err'] = ee / np.sqrt(natoms)
    thermo_info['h'] = ha / natoms
    thermo_info['h_err'] = he / np.sqrt(natoms)
    thermo_info['t'] = ta
    thermo_info['t_err'] = te
    unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
    thermo_info['pv'] = pa * va * unit_cvt / natoms
    thermo_info['pv_err'] = pe * va * unit_cvt  / np.sqrt(natoms)
    return thermo_info

def _print_thermo_info(info, more_head = '') :
    ptr = '# thermodynamics (normalized by natoms) %s\n' % more_head
    ptr += '# E (err)  [eV]:  %20.8f %20.8f\n' % (info['e'], info['e_err'])
    ptr += '# H (err)  [eV]:  %20.8f %20.8f\n' % (info['h'], info['h_err'])
    ptr += '# T (err)   [K]:  %20.8f %20.8f\n' % (info['t'], info['t_err'])
    ptr += '# P (err) [bar]:  %20.8f %20.8f\n' % (info['p'], info['p_err'])
    ptr += '# V (err) [A^3]:  %20.8f %20.8f\n' % (info['v'], info['v_err'])
    ptr += '# PV(err)  [eV]:  %20.8f %20.8f' % (info['pv'], info['pv_err'])
    print(ptr)

def _thermo_inte(jdata, Eo, Eo_err, all_t, integrand, integrand_err, scheme = 's', all_e=None) :
    path = jdata['path']
    ens = jdata['ens']
    all_temps = []
    all_press = []
    all_fe = []
    all_fe_err = []
    all_fe_sys_err = []
    
    # if all_e is not None:
    #     print('switch to test integral mode by yuanfengbo')
    #     array_e = np.asarray(all_e, dtype='float64')
    #     array_t = np.asarray(all_t, dtype='float64')
    #     array_t_delta = array_t[1:] - array_t[:-1]
    #     array_e_delta = array_e[1:] - array_e[:-1]
    #     array_b = array_e_delta / array_t_delta
    #     array_a = array_e[:-1] - array_t[:-1] * array_b 
    #     array_diff_e = array_a * (array_t[1:] - array_t[:-1])/(array_t[1:]*array_t[:-1]) + array_b * np.log(array_t[1:]/array_t[:-1])
        
    #     print('!!!', array_a, array_b, all_e , array_t_delta, array_e_delta, array_diff_e)
        
    #     # for ii in range(0, len(array_t)):
    #     for ii in range(len(array_t)-1, -1, -1):
    #         e1 = (Eo / (array_t[-1]) + np.sum(array_diff_e[ii:])) * array_t[ii]
    #         # e1 = (Eo / (array_t[0]) - np.sum(array_diff_e[0:ii])) * array_t[ii]
    #         all_temps.append(array_t[ii])
    #         err = 0
    #         sys_err = 0
    #         all_press.append(jdata['pres'])
    #         all_fe.append(e1)
    #         all_fe_err.append(err)
    #         all_fe_sys_err.append(sys_err)

    #     return np.asarray(all_temps), np.asarray(all_press), np.asarray(all_fe), np.asarray(all_fe_err), np.asarray(all_fe_sys_err)
            
            
        
    all_t, inte, inte_e, stat_e = integrate_range(all_t, integrand, integrand_err, scheme)
    for ii in range(0, len(all_t)):
        diff_e = inte[ii]
        err = stat_e[ii]
        sys_err = inte_e[ii]
        # diff_e, err = integrate(all_t[0:ii+1], integrand[0:ii+1], integrand_err[0:ii+1], scheme)
        # sys_err = integrate_sys_err(all_t[0:ii+1], integrand[0:ii+1], scheme)
        if path == 't' or path == 't-ginv':
            e1 = (Eo / (all_t[0]) - diff_e) * all_t[ii]
            err = np.sqrt(np.square(Eo_err / all_t[0]) + np.square(err))
            err *= all_t[ii]
            sys_err *= all_t[ii]
            all_temps.append(all_t[ii])
            if 'npt' in ens :
                all_press.append(get_first_matched_key_from_dict(jdata, ['pres', 'press']))
        elif path == 'p':
            e1 = Eo + diff_e        
            err = np.sqrt(np.square(Eo_err) + np.square(err))
            all_temps.append(get_first_matched_key_from_dict(jdata, ['temp', 'temps']))
            all_press.append(all_t[ii])
        all_fe.append(e1)
        all_fe_err.append(err)
        all_fe_sys_err.append(sys_err)
    return np.asarray(all_temps), np.asarray(all_press), np.asarray(all_fe), np.asarray(all_fe_err), np.asarray(all_fe_sys_err)

def post_tasks(iter_name, jdata, Eo, Eo_err = 0, To = None, natoms = None, scheme = 'simpson', shift = 0.0) :
    equi_conf = get_task_file_abspath(iter_name, jdata['equi_conf'])
    if natoms == None :        
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    ens = jdata['ens']
    path = jdata['path']

    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)
    
    all_t = []
    all_e = []
    all_e_err = []
    integrand = []
    integrand_err = []
    all_enthalpy = []
    all_msd_xyz = []
    if 'nvt' in ens and path == 't' :
        # TotEng
        stat_col = 3
        print('# TI in NVT along T path')
    elif 'npt' in ens and (path == 't' or path == 't-ginv') :
        # Enthalpy
        stat_col = 4
        print('# TI in NPT along T path')
    elif 'npt' in ens and path == 'p' :
        # volume
        stat_col = 7
        print('# TI in NPT along P path')
    else:
        raise RuntimeError('invalid ens or path setting' )
    print('# natoms: %d' % natoms)

    for ii in all_tasks :
        # get T or P
        thermo_name = os.path.join(ii, 'thermo.out')
        tt = float(open(thermo_name).read())
        all_t.append(tt)
        # get energy stat
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        ea, ee = block_avg(data[:, stat_col], 
                           skip = stat_skip, 
                           block_size = stat_bsize)
        enthalpy, _ = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
        msd_xyz = data[-1, -1]
        # COM corr
        if path == 't' or path == 't-ginv' :
            ea += 1.5 * pc.Boltzmann * tt / pc.electron_volt
            # print('~~', tt, ea, 1.5 * pc.Boltzmann * tt / pc.electron_volt)
        elif path == 'p' :
            temp = jdata['temp']
            ea += 1.5 * pc.Boltzmann * temp / pc.electron_volt
        else :
            raise RuntimeError('invalid path setting' )
        # normalized by number of atoms
        ea /= natoms
        if path == 't' or path == 't-ginv':
            ea -= shift
        ee /= np.sqrt(natoms)
        all_e.append(ea)
        all_e_err.append(ee)
        all_enthalpy.append(enthalpy)
        all_msd_xyz.append(msd_xyz)
        # gen integrand
        if path == 't' or path == 't-ginv':
            integrand.append(ea / (tt * tt))
            integrand_err.append(ee / (tt * tt))
        elif path == 'p' :
            # cvt from barA^3 to eV
            unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
            integrand.append(ea * unit_cvt)
            integrand_err.append(ee * unit_cvt)
        else:
            raise RuntimeError('invalid path setting' )

    all_print = []
    all_print.append(all_t)
    all_print.append(integrand)
    all_print.append(all_e)
    all_print.append(all_e_err)
    all_print.append(all_enthalpy)
    all_print.append(all_msd_xyz)
    all_print = np.array(all_print)
    np.savetxt(os.path.join(iter_name, 'ti.out'), 
               all_print.T, 
               fmt = '%.8e', 
               header = 't/p Integrand U/V U/V_err enthalpy msd_xyz')

    info0 = _compute_thermo(os.path.join(all_tasks[ 0], 'log.lammps'), natoms, stat_skip, stat_bsize)
    info1 = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), natoms, stat_skip, stat_bsize)
    _print_thermo_info(info0, 'at start point')
    _print_thermo_info(info1, 'at end point')

    if To is not None :
        index = all_t.index(To)
        if index == None :
            if 'nvt' == ens :
                raise RuntimeError('cannot find %f in T', To)
            elif 'npt' in ens :
                raise RuntimeError('cannot find %f in P', To)
        all_t_1 = all_t[0:index+1]
        integrand_1 = integrand[0:index+1]
        integrand_err_1 = integrand_err[0:index+1]
        all_t_1 = np.flip(all_t_1, 0)
        integrand_1 = np.flip(integrand_1, 0)
        integrand_err_1 = np.flip(integrand_err_1, 0)
        all_t_2 = all_t[index:]
        integrand_2 = integrand[index:]
        integrand_err_2 = integrand_err[index:]
        all_temps_1, all_press_1, all_fe_1, all_fe_err_1, all_fe_sys_err_1 \
            = _thermo_inte(jdata, Eo, Eo_err, all_t_1, integrand_1, integrand_err_1, scheme = scheme)
        all_temps_2, all_press_2, all_fe_2, all_fe_err_2, all_fe_sys_err_2 \
            = _thermo_inte(jdata, Eo, Eo_err, all_t_2, integrand_2, integrand_err_2, scheme = scheme)
        all_temps_1 = np.flip(all_temps_1, 0)
        all_press_1 = np.flip(all_press_1, 0)
        all_fe_1 = np.flip(all_fe_1, 0)
        all_fe_err_1 = np.flip(all_fe_err_1, 0)
        all_fe_sys_err_1 = np.flip(all_fe_sys_err_1, 0)
        all_temps = np.append(all_temps_1, all_temps_2[1:])
        all_press = np.append(all_press_1, all_press_2[1:])
        all_fe = np.append(all_fe_1, all_fe_2[1:])
        all_fe_err = np.append(all_fe_err_1, all_fe_err_2[1:])
        all_fe_sys_err = np.append(all_fe_sys_err_1, all_fe_sys_err_2[1:])
    else :    
        all_temps, all_press, all_fe, all_fe_err, all_fe_sys_err \
            = _thermo_inte(jdata, Eo, Eo_err, all_t, integrand, integrand_err, scheme = scheme, all_e=all_e)

    # print('ti.py:debug:data', data)
    result = ""
    # result_file = open(f"{iter_name}/../result", 'w')
    if 'nvt' == ens :
        print('#%8s  %20s  %9s  %9s  %9s' % ('T(ctrl)', 'F', 'stat_err', 'inte_err', 'err'))
        result += ('#%8s  %20s  %9s  %9s  %9s\n' % ('T(ctrl)', 'F', 'stat_err', 'inte_err', 'err'))
        for ii in range(len(all_temps)) :
            print ('%9.2f  %20.12f  %9.2e  %9.2e  %9.2e' 
                   % (all_temps[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii], np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]])))
            result += ('%9.2f  %20.12f  %9.2e  %9.2e  %9.2e\n' 
                   % (all_temps[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii], np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]])))
    elif 'npt' in ens :
        print('#%8s  %15s  %20s  %9s  %9s  %9s' % ('T(ctrl)', 'P(ctrl)', 'F', 'stat_err', 'inte_err', 'err'))
        result += ('#%8s  %15s  %20s  %9s  %9s  %9s\n' % ('T(ctrl)', 'P(ctrl)', 'F', 'stat_err', 'inte_err', 'err'))
        for ii in range(len(all_temps)) :
            print ('%9.2f  %15.8e  %20.12f  %9.2e  %9.2e  %9.2e' 
                   % (all_temps[ii], all_press[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii], np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]])))
            result += ('%9.2f  %15.8e  %20.12f  %9.2e  %9.2e  %9.2e\n'
                   % (all_temps[ii], all_press[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii], np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]])))
            # print(all_temps[ii], all_press[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii], np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]]))
    # result_file.close()

    data = dict(all_temps=all_temps.tolist(), all_press=all_press.tolist(),
        all_fe=all_fe.tolist(), all_fe_stat_err=all_fe_err.tolist(), all_fe_inte_err=all_fe_sys_err.tolist(), 
        all_fe_tot_err=np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]]).tolist())

    # data = [all_temps.tolist(), all_press.tolist(), 
    #     all_fe.tolist(), all_fe_err.tolist(), all_fe_sys_err.tolist(), 
    #     np.linalg.norm([all_fe_err[ii], all_fe_sys_err[ii]]).tolist()]
    info = dict(start_point_info=info0, end_point_info=info1, data=data)
    # print('result', result)
    with open(os.path.join(iter_name, '../', 'result'), 'w') as f:
        f.write(result)
    with open(os.path.join(iter_name, 'result.json'), 'w') as f:
        f.write(json.dumps(info))
    return info


def post_tasks_mbar(iter_name, jdata, Eo, natoms = None) :
    equi_conf = jdata['equi_conf']
    if natoms == None :        
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    ens = jdata['ens']
    path = jdata['path']

    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)

    if 'nvt' in ens and path == 't' :
        # TotEng
        stat_col = 3
        print('# TI in NVT along T path')
    elif 'npt' in ens and (path == 't' or path == 't-ginv') :
        # Enthalpy
        stat_col = 4
        print('# TI in NPT along T path')
    elif 'npt' in ens and path == 'p' :
        # volume
        stat_col = 7
        print('# TI in NPT along P path')
    else:
        raise RuntimeError('invalid ens or path setting' )
    print('# natoms: %d' % natoms)

    all_t = []
    for ii in all_tasks :
        thermo_name = os.path.join(ii, 'thermo.out')
        tt = float(open(thermo_name).read())
        all_t.append(tt)
    all_t = np.array(all_t)
    nt = all_t.size
    
    ukn = None
    nk = []    
    for ii in all_tasks :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        block_u = []
        if path == 't' or path == 't-ginv':
            this_e = data[stat_skip::1, stat_col]
            nk.append(this_e.size)
            for tt in all_t :
                kt_in_ev = pc.Boltzmann * tt / pc.electron_volt
                block_u.append(this_e / kt_in_ev)
        elif path == 'p' :
            this_e = data[stat_skip::1, 3]
            this_v = data[stat_skip::1, 7]
            nk.append(this_e.size)            
            # # cvt from barA^3 to eV
            unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
            temp = jdata['temps']
            kt_in_ev = temp * pc.Boltzmann / pc.electron_volt
            for tt in all_t :
                block_u.append((this_e + tt * this_v * unit_cvt) / kt_in_ev)
        else:
            raise RuntimeError('invalid path setting' )
        block_u = np.reshape(block_u, [nt, -1])
        if ukn is None :
            ukn = block_u 
        else :
            ukn = np.concatenate((ukn, block_u), axis = 1)
    nk = np.array(nk)

    info0 = _compute_thermo(os.path.join(all_tasks[ 0], 'log.lammps'), natoms, stat_skip, stat_bsize)
    info1 = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), natoms, stat_skip, stat_bsize)
    _print_thermo_info(info0, 'at start point')
    _print_thermo_info(info1, 'at end point')

    mbar = pymbar.MBAR(ukn, nk)
    Deltaf_ij, dDeltaf_ij, Theta_ij = mbar.getFreeEnergyDifferences()
    Deltaf_ij = Deltaf_ij / natoms
    dDeltaf_ij = dDeltaf_ij / np.sqrt(natoms)

    all_temps = []
    all_press = []
    all_fe = []
    all_fe_err = []
    all_fe_sys_err = []
    for ii in range(0, nt) :
        if path == 't' or path == 't-ginv':
            kt_in_ev = all_t[ii] * pc.Boltzmann / pc.electron_volt
            e1 = (Eo / (all_t[0])) * all_t[ii] + Deltaf_ij[0,ii] * kt_in_ev
            err = dDeltaf_ij[0,ii] * kt_in_ev
            sys_err = 0
            all_temps.append(all_t[ii])
            if 'npt' in ens :
                all_press.append(jdata['pres'])
        elif path == 'p':
            kt_in_ev = jdata['temps'] * pc.Boltzmann / pc.electron_volt
            e1 = Eo + Deltaf_ij[0,ii] * kt_in_ev
            err = dDeltaf_ij[0,ii] * kt_in_ev
            sys_err = 0
            all_temps.append(jdata['temp'])
            all_press.append(all_t[ii])            
        else :
            pass
        all_fe.append(e1)
        all_fe_err.append(err)
        all_fe_sys_err.append(sys_err)

    if 'nvt' == ens :
        print('#%8s  %15s  %9s  %9s' % ('T(ctrl)', 'F', 'stat_err', 'inte_err'))
        for ii in range(len(all_temps)) :
            print ('%9.2f  %20.12f  %9.2e  %9.2e' 
                   % (all_temps[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii]))
    elif 'npt' in ens :
        print('#%8s  %15s  %15s  %9s  %9s' % ('T(ctrl)', 'P(ctrl)', 'F', 'stat_err', 'inte_err'))
        for ii in range(len(all_temps)) :
            print ('%9.2f  %15.8e  %20.12f  %9.2e  %9.2e' 
                   % (all_temps[ii], all_press[ii], all_fe[ii], all_fe_err[ii], all_fe_sys_err[ii]))
    # info = dict(start_point_info=info0, end_point_info=info1, all_temps=list(all_temps), all_press=list(all_press),
    #              all_fe=list(all_fe), all_fe_err=list(all_fe_err), all_fe_sys_err=list(all_fe_sys_err))
    # open(os.path.join(iter_name, 'result.json'), 'w').write(json.dumps(info))
    # return info


def refine_task (from_task, to_task, err) :
    from_task = os.path.abspath(from_task)
    to_task = os.path.abspath(to_task)
    from_json = os.path.join(from_task, 'in.json')
    to_json = os.path.join(to_task, 'in.json')
    from_jdata = json.load(open(from_json))
    to_jdata = from_jdata
    path = from_jdata['path']

    from_ti = os.path.join(from_task, 'ti.out')
    if not os.path.isfile(from_ti) :
        raise RuntimeError("cannot find file %s, task should be computed befor refined" % from_ti)
    tmp_array = np.loadtxt(from_ti)
    all_t = tmp_array[:,0]
    integrand = tmp_array[:,1]
    ntask = all_t.size

    if path == 't' or path == 't-ginv':
        interval_nrefine = compute_nrefine(all_t, integrand, err, all_t)
    elif path == 'p' :
        interval_nrefine = compute_nrefine(all_t, integrand, err)
    else :
        raise RuntimeError('unknow path ' + path)

    refined_t = []
    back_map = []
    for ii in range(0, ntask-1) :
        refined_t.append(all_t[ii])
        back_map.append(ii)
        hh = (all_t[ii+1] - all_t[ii]) / interval_nrefine[ii]
        for jj in range(1, interval_nrefine[ii]) :
            refined_t.append(all_t[ii] + jj * hh)
            back_map.append(-1)
    refined_t.append(all_t[-1])
    back_map.append(ntask-1)
    
    if to_jdata['path'] == 't-ginv' :
        to_jdata['path'] = 't'        
    if to_jdata['path'] == 't' :
        to_jdata['temps'] = refined_t
    elif to_jdata['path'] == 'p' :
        to_jdata['press'] = refined_t
    else :
        raise RuntimeError('unknow path ' + path)
    to_jdata['orig_task'] = from_task
    to_jdata['back_map'] = back_map
    to_jdata['refine_error'] = err
    to_jdata['equi_conf'] = get_task_file_abspath(from_task, from_jdata['equi_conf'])
    to_jdata['model'] = get_task_file_abspath(from_task, from_jdata['model'])
    # create_path(to_task)
    # with open(to_json, 'w') as fp :
    #     json.dump(to_jdata, fp, indent=4)

    make_tasks(to_task, to_jdata)
    
    from_task_list = glob.glob(os.path.join(from_task, 'task.[0-9]*'))
    from_task_list.sort()
    to_task_list = glob.glob(os.path.join(to_task, 'task.[0-9]*'))
    to_task_list.sort()
    assert(len(from_task_list) == ntask)
    assert(len(to_task_list) == len(refined_t))

    for ii in range(len(to_task_list)) :
        if back_map[ii] < 0 : 
            continue
        for jj in ['data', 'log.lammps'] :
            shutil.copyfile(
                os.path.join(from_task_list[back_map[ii]], jj), 
                os.path.join(to_task_list[ii], jj), 
            )
        with open(os.path.join(to_task_list[ii], 'from.dir'), 'w') as fp:
            fp.write(from_task_list[back_map[ii]])
            
def compute_task(job, inte_method, Eo, Eo_err, To, scheme='simpson'):
    # job = args.JOB
    with open(os.path.join(job, 'ti_settings.json'), 'r') as f:
        jdata = json.load(f)
    if inte_method == 'inte' :
        info = post_tasks(job, jdata, Eo=Eo, Eo_err=Eo_err, To=To, scheme=scheme)
    elif inte_method == 'mbar' :
        info = post_tasks_mbar(job, jdata, Eo)
    else :
        raise RuntimeError('unknow integration method')
    return info
    

def _main ():
    parser = argparse.ArgumentParser(
        description="Compute free energy by TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command', help = 'valid commands')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')
    parser_gen.add_argument("-z", "--meam", help="whether use meam instead of dp", action="store_true")

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-m','--inte-method', type=str, default = 'inte', 
                             choices=['inte', 'mbar'], 
                             help='the method of thermodynamic integration')
    parser_comp.add_argument('-e', '--Eo', type=float, default = 0,
                             help='free energy of starting point')
    parser_comp.add_argument('-E', '--Eo-err', type=float, default = 0,
                             help='The statistical error of the starting free energy')
    parser_comp.add_argument('-t', '--To', type=float, 
                             help='the starting thermodynamic position')
    parser_comp.add_argument('-s', '--scheme', type=str, default = 'simpson',
                             help='the numerical integration scheme')

    parser_comp = subparsers.add_parser('refine', help= 'Refine the grid of a job')
    parser_comp.add_argument('-i', '--input', type=str, required=True,
                             help='input job')
    parser_comp.add_argument('-o', '--output', type=str, required=True,
                             help='output job')
    parser_comp.add_argument('-e', '--error', type=float, required=True,
                             help='the error required')
    args = parser.parse_args()

    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        jdata = json.load(open(args.PARAM, 'r'))
        make_tasks(output, jdata, if_meam=args.meam)
    elif args.command == 'compute' :
        compute_task(args.JOB, inte_method=args.inte_method, Eo=args.Eo, Eo_err=args.Eo_err, To=args.To, scheme=args.scheme)
    #     job = args.JOB
    #     jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))
    #     if args.inte_method == 'inte' :
    #         post_tasks(job, jdata, args.Eo, Eo_err = args.Eo_err, To = args.To, scheme = args.scheme)
    #     elif args.inte_method == 'mbar' :
    #         post_tasks_mbar(job, jdata, args.Eo)
    #     else :
    #         raise RuntimeError('unknow integration method')
    elif args.command == 'refine' :
        refine_task(args.input, args.output, args.error)

    
if __name__ == '__main__' :
    _main()
        



import json, os, sys, glob

from datetime import datetime, timedelta

import aiida
from aiida.engine import calcfunction, workfunction
from aiida.orm import Int, Dict, Str

# from dpdispatcher.lazy_local_context import LazyLocalContext
from dpdispatcher.submission import Submission, Machine, Task, Resources
from dpti import equi, hti, hti_liq, ti
import subprocess as sp

aiida.load_profile()

def get_empty_submission(job_work_dir):
    machine_file = os.path.join(job_work_dir, '../', '../', '../', 'machine.json')
    with open(machine_file, 'r') as f:
        mdata = json.load(f)

    machine = Machine.load_from_dict(mdata['machine'])
    resources = Resources.load_from_dict(mdata['resources'])

    submission = Submission(
        work_base=job_work_dir, 
        resources=resources, 
        machine=machine, 
    )
    return submission

@calcfunction
def all_start_check(dag_run):
    work_base_dir = dag_run['work_base_dir']
    target_temp = int(dag_run['target_temp'])
    target_pres = int(dag_run['target_pres'])
    conf_lmp = str(dag_run['conf_lmp'])
    ti_path = str(dag_run['ti_path'])
    ens = str(dag_run['ens'])
    if_liquid = dag_run['if_liquid']

    work_base_abs_dir = os.path.realpath(work_base_dir)

    dag_work_dirname=str(target_temp)+'K-'+str(target_pres)+'bar-'+str(conf_lmp)
    dag_work_dir=os.path.join(work_base_abs_dir, dag_work_dirname)
    
    assert os.path.isdir(work_base_dir) is True,  f'work_base_dir {work_base_dir} must exist '
    if os.path.isdir(dag_work_dir) is False:
        os.mkdir(dag_work_dir)
    else:
        pass

    conf_lmp_abs_path = os.path.join(work_base_abs_dir, conf_lmp)
    assert os.path.isfile(conf_lmp_abs_path) is True,  f'structure file {conf_lmp_abs_path} must exist'
    assert str(ti_path) in ["t", "p"], f'value for "path" must be "t" or "p" '
    start_info = dict(work_base_dir=work_base_dir, 
        target_temp=target_temp,
        target_pres=target_pres, 
        conf_lmp=conf_lmp, 
        ti_path=ti_path,
        ens=ens, 
        if_liquid=if_liquid, 
        work_base_abs_dir=work_base_abs_dir,
        dag_work_dir=dag_work_dir)
    # print('start_info:', start_info_raw)
    # start_info = Dict(dict=start_info)
    return Dict(dict=start_info)

@calcfunction
def NPT_start(start_info):
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']
    job_work_dir = Str(os.path.join(dag_work_dir, 'NPT_sim', 'new_job'))

    result_json_file = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_json_file):
        return job_work_dir
    with open(os.path.join(work_base_abs_dir, 'npt.json')) as f:
        npt_jdata = json.load(f)

    task_jdata = npt_jdata.copy()
    task_jdata['equi_conf'] = start_info['conf_lmp']
    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']
    task_jdata['ens'] = start_info['ens']
    print(task_jdata)
    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    equi.make_task(job_work_dir.value, task_jdata)
    os.chdir(cwd)
    return job_work_dir

@calcfunction
def NPT_sim(job_work_dir):
    task = Task(command='lmp -i in.lammps', task_work_path='./')
    submission = get_empty_submission(job_work_dir.value)
    submission.register_task_list([task])
    submission.run_submission()
    return Str(job_work_dir.value)

@calcfunction
def NPT_end(job_work_dir):
    result_file_path = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_file_path):
        info = json.load(open(result_file_path, 'r'))
    else:
        info = equi.post_task(job_work_dir.value)
    print(info)
    # info = Dict(dict=info)
    return Dict(dict=info)


@calcfunction
def NVT_start(start_info, NPT_end_info):
    npt_dir = NPT_end_info['job_dir']
    
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']
    job_work_dir = Str(os.path.join(dag_work_dir, 'NVT_sim', 'new_job'))

    result_json_file = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_json_file):
        return job_work_dir

    with open(os.path.join(work_base_abs_dir, 'nvt.json')) as f:
        nvt_jdata = json.load(f)

    task_jdata = nvt_jdata.copy()
    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']
    task_jdata['ens'] = 'nvt' 
    
    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    equi.make_task(job_work_dir.value, task_jdata, npt_dir=npt_dir)
    os.chdir(cwd)
    return job_work_dir

@calcfunction
def NVT_sim(job_work_dir):
    submission = get_empty_submission(job_work_dir.value)
    task = Task(command='lmp -i in.lammps', task_work_path='./') 
    submission.register_task_list([task])
    submission.run_submission()
    return Str(job_work_dir.value)

@calcfunction
def NVT_end(job_work_dir):
    result_file_path = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_file_path):
        with open(result_file_path, 'r') as f:
            info = json.load(f)
    else:
        info = equi.post_task(job_work_dir.value)
    print(info)
    return Dict(dict=info)


@calcfunction
def HTI_start(start_info, NVT_end_info):
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']
    if_liquid = start_info['if_liquid']
    conf_lmp = NVT_end_info['out_lmp']

    job_work_dir = Str(os.path.join(dag_work_dir, 'HTI_sim', 'new_job'))
    if os.path.isfile(os.path.join(dag_work_dir, 'HTI_sim', 'new_job', 'result.json')):
        return job_work_dir

    if if_liquid:
        with open(os.path.join(work_base_abs_dir, 'hti.liquid.json')) as j:
            hti_jdata = json.load(j)
    else:
        with open(os.path.join(work_base_abs_dir, 'hti.json')) as j:
            hti_jdata = json.load(j)
    
    task_jdata = hti_jdata.copy()
    # task_jdata['equi_conf'] = "../NVT_sim/new_job/out.lmp"
    # task_jdata['equi_conf'] = os.path.join(start_info['dag_work_dir'], 'NVT_sim', 'new_job', 'out.lmp')
    if conf_lmp is not None:
        task_jdata['equi_conf'] = conf_lmp
    else:
        task_jdata['equi_conf'] = start_info['conf_lmp']

    task_jdata['temp'] = start_info['target_temp']
    task_jdata['pres'] = start_info['target_pres']

    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    if if_liquid:
        hti_liq.make_tasks(iter_name=job_work_dir.value, jdata=task_jdata)
    else:
        hti.make_tasks(iter_name=job_work_dir.value, jdata=task_jdata, ref='einstein', switch='three-step')
    os.chdir(cwd)
    return job_work_dir

@calcfunction
def HTI_sim(job_work_dir):
    task_abs_dir_list = glob.glob(os.path.join(job_work_dir.value, './*/task*'))
    task_dir_list = [os.path.relpath(ii, start=job_work_dir.value ) for ii in task_abs_dir_list]

    task_list = [ Task(command='lmp_serial -i in.lammps', 
        task_work_path=ii) 
        for ii in task_dir_list ]
    submission = get_empty_submission(job_work_dir.value)
    submission.register_task_list(task_list=task_list)
    submission.run_submission()
    return Str(job_work_dir.value)

@calcfunction
def HTI_end(job_work_dir,
    start_info,
    NPT_end_info
):
    if_liquid = start_info['if_liquid']
    manual_pv = NPT_end_info['pv']
    manual_pv_err = NPT_end_info['pv_err']

    result_file_path = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_file_path):
        info = json.load(open(result_file_path, 'r'))
        return info

    if if_liquid:
        info = hti_liq.compute_task(
            job_work_dir.value, 
            'gibbs', 
            manual_pv=manual_pv, 
            manual_pv_err=manual_pv_err
        )
    else:
        info = hti.compute_task(
            job_work_dir.value, 
            'gibbs', 
            manual_pv=manual_pv, 
            manual_pv_err=manual_pv_err
        )
    print(info)
    return Dict(dict=info)

@calcfunction
def TI_start(start_info, HTI_end_info):
    work_base_abs_dir = start_info['work_base_abs_dir']
    dag_work_dir = start_info['dag_work_dir']

    ti_path = start_info['ti_path']
    conf_lmp = start_info['conf_lmp']

    if ti_path == 't':
        with open(os.path.join(work_base_abs_dir, 'ti.t.json')) as j:
            ti_jdata = json.load(j)
            task_jdata = ti_jdata.copy()
            task_jdata['pres'] = start_info['target_pres']
            job_dir = 'TI_t_sim'
    elif ti_path == 'p':
        with open(os.path.join(work_base_abs_dir, 'ti.p.json')) as j:
            ti_jdata = json.load(j)
            task_jdata = ti_jdata.copy()
            task_jdata['temp'] = start_info['target_temp']
            job_dir = 'TI_p_sim'
    else:
        raise RuntimeError(f'Error integration path. {ti_path}')

    job_work_dir = Str(os.path.join(dag_work_dir, job_dir, 'new_job'))
    task_jdata['ens'] = start_info['ens']
    task_jdata['equi_conf'] = conf_lmp

    cwd = os.getcwd()
    os.chdir(work_base_abs_dir)
    ti.make_tasks(job_work_dir.value, task_jdata)
    os.chdir(cwd)
    return job_work_dir

@calcfunction
def TI_sim(job_work_dir):
    task_abs_dir_list = glob.glob(os.path.join(job_work_dir.value, './task*'))
    task_dir_list = [os.path.relpath(ii, start=job_work_dir.value ) for ii in task_abs_dir_list]
    submission = get_empty_submission(job_work_dir.value)
    task_list = [ Task(command='lmp_serial -i in.lammps', task_work_path=ii) for ii in task_dir_list ]
    submission.register_task_list(task_list=task_list)
    submission.run_submission()
    return Str(job_work_dir.value)

@calcfunction
def TI_end(job_work_dir, start_info, HTI_end_info):
    Eo = HTI_end_info['e1']
    Eo_err = HTI_end_info['e1_err']
    # ti_jdata = json.load(open(os.path.join(sim_work_dir, 'in.json'), 'r'))

    ti_path = start_info['ti_path']
    if ti_path == 't':
        To = start_info['target_temp']
    if ti_path == 'p':
        To = start_info['target_pres']
    result_file_path = os.path.join(job_work_dir.value, 'result.json')
    if os.path.isfile(result_file_path):
        with open(result_file_path, 'r') as f:
            info = json.load(f)
    else:
        info = ti.compute_task(job_work_dir.value, inte_method='inte', Eo=Eo, Eo_err=Eo_err, To=To)
    print(info)
    return Dict(dict=info)


@workfunction
def TI_workflow(dag_run):
    start_info = all_start_check(dag_run)
    npt_job_dir = NPT_sim(NPT_start(start_info=start_info))
    NPT_end_info = NPT_end(npt_job_dir)
    
    nvt_job_dir = NVT_sim(NVT_start(start_info=start_info, NPT_end_info=NPT_end_info))
    NVT_end_info = NVT_end(nvt_job_dir)

    hti_job_dir = HTI_sim(HTI_start(
        start_info=start_info,
        NVT_end_info=NVT_end_info))
    HTI_end_info = HTI_end(hti_job_dir, start_info, NPT_end_info)

    ti_end_info = TI_sim(TI_start(
        start_info=start_info,
        HTI_end_info=HTI_end_info))
    TI_end_info = TI_end(ti_end_info, start_info, HTI_end_info)

    return TI_end_info

with open("../examples/FreeEnergy.json", "r") as f:
    jdata = json.load(f)

print(jdata)
dag_run = Dict(dict=jdata)
result = TI_workflow(dag_run=dag_run)
print(result)




#!/usr/bin/env python3

import os, sys, json, argparse, glob
import numpy as np

from lib.utils import create_path
from lib.utils import cvt_conf
from lib.lammps import get_last_dump

def _gen_lammps_relax (conf_file, 
                       mass_map, 
                       model, 
                       pres, 
                       thermo_freq = 100, 
                       dump_freq = 100) :
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        THERMO_FREQ     equal %d\n' % thermo_freq
    ret += 'variable        DUMP_FREQ       equal %d\n' % dump_freq
    ret += 'variable        PRES            equal %f\n' % pres
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      atomic\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % conf_file
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)) :
        ret+= "mass            %d %f\n" %(jj+1, mass_map[jj])
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    ret += 'pair_style      deepmd %s\n' % model
    ret += 'pair_coeff\n'
    ret += '# --------------------- MD SETTINGS ----------------------\n'
    ret += 'neighbor        1.0 bin\n'
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'thermo_style    custom step pe enthalpy press vol pxx pyy pzz pxy pyz pxz\n'
    ret += 'dump		1 all custom 100 dump.relax id type x y z vx vy vz fx fy fz\n'
    ret += '# --------------------- RUN ------------------------------\n'
    ret += 'min_style       cg\n'
    ret += 'fix             1 all box/relax iso ${PRES}\n'
    ret += 'minimize        1.000000e-12 1.000000e-12 500000 500000\n'
    ret += 'fix             1 all box/relax tri ${PRES}\n'
    ret += 'minimize        1.000000e-12 1.000000e-12 500000 500000\n'
    return ret

def make_task(iter_name, jdata, pres) :
    equi_conf = jdata['equi_conf']
    equi_conf = os.path.abspath(equi_conf)
    model = jdata['model']
    model = os.path.abspath(model)
    model_mass_map = jdata['model_mass_map']
    if pres == None :
        pres = jdata['pres']
    elif 'pres' in jdata :
        print('P = %f overrides the pres in json data' % pres)
    jdata['pres'] = pres
    
    create_path(iter_name)
    cwd = os.getcwd()
    os.chdir(iter_name)
    with open('in.json', 'w') as fp:
        json.dump(jdata, fp, indent=4)
    os.symlink(os.path.relpath(equi_conf), 'conf.lmp')
    os.symlink(os.path.relpath(model), 'graph.pb')
    lmp_str \
        = _gen_lammps_relax('conf.lmp',
                            model_mass_map, 
                            'graph.pb',
                            pres)
    with open('in.lammps', 'w') as fp :
        fp.write(lmp_str)
    os.chdir(cwd)

def extract(iter_name, output) :
    if os.path.exists(output) :
        raise RuntimeError('existing file ' + output + ' do nothing')
    dump_file = os.path.join(iter_name, 'dump.relax')
    with open('dump.tmp', 'w') as fp :
        fp.write(get_last_dump(dump_file))
    cvt_conf('dump.tmp', output, ofmt = 'lammps_data')
    os.remove('dump.tmp')

def compute(iter_name) :
    fname = os.path.join(iter_name, 'log.lammps')
    lines = open(fname).read().split('\n')
    for ii in range(len(lines)) :
        if 'Loop time of' in lines[ii] :
            idx = ii-1
            break
    res = [float(jj) for jj in lines[idx].split()]
    return res[1], res[2]

def _main ():
    parser = argparse.ArgumentParser(
        description="Relax conf")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-p','--pressure', type=float,
                            help='the pressure of the system')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')

    parser_comp = subparsers.add_parser('extract', help= 'Extract the conf')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-o','--output', type=str, default = 'conf.lmp',
                             help='output conf file name')

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-t','--type', type=str, default = 'helmholtz', 
                             choices=['helmholtz', 'gibbs'], 
                             help='the type of free energy')

    args = parser.parse_args()

    
    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        jdata = json.load(open(args.PARAM, 'r'))        
        make_task(args.output, jdata, args.pressure)
    elif args.command == 'extract' :
        extract(args.JOB, args.output)
    elif args.command == 'compute' :
        ener, enthalpy = compute(args.JOB)
        if args.type == 'helmholtz' :
            print('# Helmholtz free ener (err) [eV] at 0K == energy:')
            print(ener)
        if args.type == 'gibbs' :
            print('# Gibbs free ener (err) [eV] at 0K == ethalpy:')
            print(enthalpy)
        

if __name__ == '__main__' :
    _main()



#!/usr/bin/env python3

import os, sys, json, argparse, glob, shutil
import numpy as np
import scipy.constants as pc

# sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../'))
from dpti import einstein
import dpti.lib.lmp as lmp
from dpti.lib.utils import create_path, relative_link_file
from dpti.lib.utils import copy_file_list
from dpti.lib.utils import block_avg
from dpti.lib.utils import integrate
from dpti.lib.utils import integrate_sys_err
from dpti.lib.utils import parse_seq
from dpti.lib.utils import get_first_matched_key_from_dict
from dpti.lib.lammps import get_thermo
from dpti.lib.lammps import get_natoms

def make_iter_name (iter_index) :
    return "task_hti." + ('%04d' % iter_index)

def _ff_soft_on(lamb, 
                sparam) :
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'pair_style      lj/cut/soft %f %f %f\n' % (nn, alpha_lj, rcut)

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * ${EPSILON} %f %f\n' % (sigma, activation)
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON\n'
    return ret

def _ff_deep_on(lamb, 
                sparam, 
                model,
                if_meam=False,
                meam_model=None):
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        ONE equal 1\n'
    if if_meam:
        ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f\n' % (nn, alpha_lj, rcut)
        ret += f'pair_coeff      * * meam {meam_model["library"]} {meam_model["element"]} {meam_model["potential"]} {meam_model["element"]}\n'
        # ret += f'pair_coeff      * * meam {meam_model[0]} {meam_model[2]} {meam_model[1]} {meam_model[2]}\n'
    else:
        ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f\n' % (model, nn, alpha_lj, rcut)
        ret += 'pair_coeff      * * deepmd\n'

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    if if_meam:
        ret += 'fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA\n'
        ret += 'compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE\n'
    else:
        ret += 'fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA\n'
        ret += 'compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE\n'
    return ret

def _ff_soft_off(lamb, 
                 sparam, 
                 model,
                 if_meam=False,
                 meam_model=None) :
    nn = sparam['n']
    alpha_lj = sparam['alpha_lj']
    rcut = sparam['rcut']
    epsilon = sparam['epsilon']
    # sigma = sparam['sigma']
    activation = sparam['activation']
    ret = ''
    ret += 'variable        INV_LAMBDA equal 1-${LAMBDA}\n'
    ret += 'variable        EPSILON equal %f\n' % epsilon
    ret += 'variable        INV_EPSILON equal -${EPSILON}\n'
    if if_meam:
        ret += 'pair_style      hybrid/overlay meam lj/cut/soft %f %f %f\n' % (nn, alpha_lj, rcut)
        ret += f'pair_coeff      * * meam {meam_model["library"]} {meam_model["element"]} {meam_model["potential"]} {meam_model["element"]}\n'
        # ret += f'pair_coeff      * * meam {meam_model[0]} {meam_model[2]} {meam_model[1} {meam_model[2]} \n'
    else:
        ret += 'pair_style      hybrid/overlay deepmd %s lj/cut/soft %f %f %f\n' % (model, nn, alpha_lj, rcut)
        ret += 'pair_coeff      * * deepmd\n'

    element_num=sparam.get('element_num', 1)
    sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
    for (i, j) in sigma_key_index:
        ret += 'pair_coeff      %s %s lj/cut/soft ${EPSILON} %f %f\n' % (i+1, j+1, sparam['sigma_'+str(i)+'_'+str(j)], activation)

    # ret += 'pair_coeff      * * lj/cut/soft ${EPSILON} %f %f\n' % (sigma, activation)
    ret += 'fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes\n'
    ret += 'compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON\n'
    return ret

def _gen_lammps_input_ideal (step,
                             conf_file, 
                             mass_map,
                             lamb,
                             soft_param,
                             model,
                             nsteps,
                             timestep,
                             ens,
                             temp,
                             pres = 1.0, 
                             tau_t = 0.1,
                             tau_p = 0.5,
                             thermo_freq = 100, 
                             copies = None,
                             norm_style = 'first',
                             if_meam = False,
                             meam_model = None) :
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        NSTEPS          equal %d\n' % nsteps
    ret += 'variable        THERMO_FREQ     equal %d\n' % thermo_freq
    # ret += 'variable        DUMP_FREQ       equal %d\n' % thermo_freq
    ret += 'variable        TEMP            equal %f\n' % temp
    ret += 'variable        PRES            equal %f\n' % pres
    ret += 'variable        TAU_T           equal %f\n' % tau_t
    ret += 'variable        TAU_P           equal %f\n' % tau_p
    ret += 'variable        LAMBDA          equal %.10e\n' % lamb
    ret += 'variable        ZERO            equal 0\n'
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      atomic\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % conf_file
    if copies is not None :
        ret += 'replicate       %d %d %d\n' % (copies[0], copies[1], copies[2])
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)) :
        ret += "mass            %d %f\n" %(jj+1, mass_map[jj])
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    if step == 'soft_on' :
        ret += _ff_soft_on(lamb, soft_param)
    elif step == 'deep_on' :
        ret += _ff_deep_on(lamb, soft_param, model, if_meam=if_meam, meam_model=meam_model)
    elif step == 'soft_off' :
        ret += _ff_soft_off(lamb, soft_param, model, if_meam=if_meam, meam_model=meam_model)
    else :
        raise RuntimeError('unknown step')
    ret += '# --------------------- MD SETTINGS ----------------------\n'    
    ret += 'neighbor        1.0 bin\n'
    ret += 'timestep        %s\n' % timestep
    ret += 'compute         allmsd all msd\n'
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol c_e_diff[1] c_allmsd[*]\n'
    ret += 'thermo_modify   format 9 %.16e\n'
    ret += '# dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz\n'
    if ens == 'nvt' :
        ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
    elif ens == 'npt-iso' or ens == 'npt':
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'nve' :
        ret += 'fix             1 all nve\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)        
    ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
    ret += '# --------------------- INITIALIZE -----------------------\n'    
    ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
    ret += 'velocity        all zero linear\n'
    ret += '# --------------------- RUN ------------------------------\n'    
    ret += 'run             ${NSTEPS}\n'
    ret += 'write_data      out.lmp\n'
    
    return ret


def _make_tasks(iter_name, jdata, step, if_meam=False, meam_model=None) :
    if step == 'soft_on' :
        all_lambda = parse_seq(jdata['lambda_soft_on'])
    elif step == 'deep_on' :
        all_lambda = parse_seq(jdata['lambda_deep_on'])
    elif step == 'soft_off' :
        all_lambda = parse_seq(jdata['lambda_soft_off'])
    else :
        raise RuntimeError('unknow step')
    equi_conf = jdata['equi_conf']
    equi_conf = os.path.abspath(equi_conf)
    # mass_map = jdata['mass_map']
    mass_map = get_first_matched_key_from_dict(jdata, ['mass_map', 'model_mass_map'])
    model = jdata.get('model', None)
    if model:
        model = os.path.abspath(model)
    soft_param = jdata['soft_param']
    nsteps = jdata['nsteps']
    # timestep = jdata['timestep']
    timestep = get_first_matched_key_from_dict(jdata, ['timestep', 'dt'])
    # thermo_freq = jdata['thermo_freq']
    thermo_freq = get_first_matched_key_from_dict(jdata, ['thermo_freq', 'stat_freq'])
    copies = None
    if 'copies' in jdata :
        copies = jdata['copies']
    temp = jdata['temp']
    
    sparam = jdata.get('soft_param', {})
    if sparam:
        if 'sigma_oo' in sparam:
            sparam['sigma_0_0'] = sparam['sigma_oo']
            sparam['sigma_0_1'] = sparam['sigma_oh']
            sparam['sigma_1_1'] = sparam['sigma_hh']

        # element_num=sparam.get('element_num', 1)
        element_num = len(mass_map)
        sigma_key_index = filter(lambda t:t[0] <= t[1], ((i,j) for i in range(element_num) for j in range(element_num)))
        sigma_key_name_list = ['sigma_'+str(t[0])+'_'+str(t[1]) for t in sigma_key_index ]
        for sigma_key_name in sigma_key_name_list:
            assert sparam.get(sigma_key_name, None), 'there must be key-value for {sigma_key_name} in soft_param'.format(sigma_key_name=sigma_key_name)

    job_abs_dir = create_path(iter_name)

    if meam_model:
        relative_link_file(os.path.abspath(meam_model['library']), job_abs_dir)
        relative_link_file(os.path.abspath(meam_model['potential']), job_abs_dir)
        # os.symlink(os.path.join('..', 'conf.lmp'), 'conf.lmp')
        # os.symlink(os.path.join('..', 'graph.pb'), 'graph.pb')

    cwd = os.getcwd()
    os.chdir(iter_name)
    os.symlink(os.path.join('..', 'in.json'), 'in.json')
    os.symlink(os.path.join('..', 'conf.lmp'), 'conf.lmp')
    os.symlink(os.path.join('..', 'graph.pb'), 'graph.pb')

    os.chdir(cwd)
    # print(9898, meam_model)
    for idx,ii in enumerate(all_lambda) :
        work_path = os.path.join(iter_name, 'task.%06d' % idx)
        create_path(work_path)
        os.chdir(work_path)
        os.symlink(os.path.join('..', 'conf.lmp'), 'conf.lmp')
        os.symlink(os.path.join('..', 'graph.pb'), 'graph.pb')
        if meam_model:
            meam_library_basename = os.path.basename(meam_model['library'])
            meam_potential_basename = os.path.basename(meam_model['potential'])
            os.symlink(os.path.join('..', meam_library_basename), meam_library_basename)
            os.symlink(os.path.join('..', meam_potential_basename), meam_potential_basename)
        lmp_str \
            = _gen_lammps_input_ideal(step, 
                                      'conf.lmp',
                                      mass_map, 
                                      ii, 
                                      soft_param,
                                      'graph.pb',
                                      nsteps, 
                                      timestep,
                                      'nvt',
                                      temp,
                                      thermo_freq = thermo_freq, 
                                      copies = copies,
                                      if_meam = if_meam,
                                      meam_model=meam_model)
        with open('in.lammps', 'w') as fp :
            fp.write(lmp_str)
        with open('lambda.out', 'w') as fp :
            fp.write(str(ii))
        os.chdir(cwd)


def make_tasks(iter_name, jdata, if_meam=None) :
    if not if_meam :
        if_meam = jdata['if_meam']
    equi_conf = os.path.abspath(jdata['equi_conf'])
    if jdata.get('model', None):
        model = os.path.abspath(jdata['model'])
    else:
        model = None
    meam_model = jdata.get('meam_model', None)

    create_path(iter_name)
    copied_conf = os.path.join(os.path.abspath(iter_name), 'conf.lmp')
    shutil.copyfile(equi_conf, copied_conf)
    # jdata['equi_conf'] = copied_conf
    if model:
        copied_model = os.path.join(os.path.abspath(iter_name), 'graph.pb')
        shutil.copyfile(model, copied_model)
    # jdata['model'] = copied_model

    cwd = os.getcwd()
    os.chdir(iter_name)    
    with open('in.json', 'w') as fp:
        json.dump(jdata, fp, indent=4)
    os.chdir(cwd)
    subtask_name = os.path.join(iter_name, '00.soft_on')
    _make_tasks(subtask_name, jdata, 'soft_on', if_meam=if_meam, meam_model=meam_model)
    subtask_name = os.path.join(iter_name, '01.deep_on')
    _make_tasks(subtask_name, jdata, 'deep_on', if_meam=if_meam, meam_model=meam_model)
    subtask_name = os.path.join(iter_name, '02.soft_off')
    _make_tasks(subtask_name, jdata, 'soft_off', if_meam=if_meam, meam_model=meam_model)


def _compute_thermo(fname, natoms, stat_skip, stat_bsize) :
    data = get_thermo(fname)
    ea, ee = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
    ha, he = block_avg(data[:, 4], skip = stat_skip, block_size = stat_bsize)
    ta, te = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
    pa, pe = block_avg(data[:, 6], skip = stat_skip, block_size = stat_bsize)
    va, ve = block_avg(data[:, 7], skip = stat_skip, block_size = stat_bsize)
    thermo_info = {}
    thermo_info['p'] = pa
    thermo_info['p_err'] = pe
    thermo_info['v'] = va / natoms
    thermo_info['v_err'] = ve / np.sqrt(natoms)
    thermo_info['e'] = ea / natoms
    thermo_info['e_err'] = ee / np.sqrt(natoms)
    thermo_info['h'] = ha / natoms
    thermo_info['h_err'] = he / np.sqrt(natoms)
    thermo_info['t'] = ta
    thermo_info['t_err'] = te
    unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
    thermo_info['pv'] = pa * va * unit_cvt / natoms
    thermo_info['pv_err'] = pe * va * unit_cvt  / np.sqrt(natoms)
    return thermo_info


def _post_tasks(iter_name, step, natoms) :
    jdata = json.load(open(os.path.join(iter_name, 'in.json')))
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    all_tasks = glob.glob(os.path.join(iter_name, 'task.[0-9]*'))
    all_tasks.sort()
    ntasks = len(all_tasks)
    
    all_lambda = []
    all_dp_a = []
    all_dp_e = []
    all_msd_xyz = []

    for ii in all_tasks :
        log_name = os.path.join(ii, 'log.lammps')
        data = get_thermo(log_name)
        np.savetxt(os.path.join(ii, 'data'), data, fmt = '%.6e')
        dp_a, dp_e = block_avg(data[:, 8], skip = stat_skip, block_size = stat_bsize)
        msd_xyz = data[-1, 12]
        dp_a /= natoms
        dp_e /= np.sqrt(natoms)
        lmda_name = os.path.join(ii, 'lambda.out')
        ll = float(open(lmda_name).read())
        all_lambda.append(ll)
        all_dp_a.append(dp_a)
        all_dp_e.append(dp_e)
        all_msd_xyz.append(msd_xyz)

    all_lambda = np.array(all_lambda)
    all_dp_a = np.array(all_dp_a)
    all_dp_e = np.array(all_dp_e)
    all_msd_xyz = np.array(all_msd_xyz)
    de = all_dp_a
    all_err = all_dp_e

    all_print = []
    # all_print.append(np.arange(len(all_lambda)))
    all_print.append(all_lambda)
    all_print.append(de)
    all_print.append(all_err)
    all_print.append(all_msd_xyz)
    all_print = np.array(all_print)
    np.savetxt(os.path.join(iter_name, 'hti.out'), 
               all_print.T, 
               fmt = '%.8e', 
               header = 'lmbda dU dU_err msd_xyz')

    diff_e, err = integrate(all_lambda, de, all_err)
    sys_err = integrate_sys_err(all_lambda, de)

    thermo_info = _compute_thermo(os.path.join(all_tasks[-1], 'log.lammps'), 
                                  natoms,
                                  stat_skip, stat_bsize)

    return diff_e, [err, sys_err], thermo_info


def post_tasks(iter_name, natoms) :
    fe = einstein.ideal_gas_fe(iter_name)
    subtask_name = os.path.join(iter_name, '00.soft_on')
    e0, err0, tinfo0 = _post_tasks(subtask_name, 'soft_on', natoms)
    subtask_name = os.path.join(iter_name, '01.deep_on')
    e1, err1, tinfo1 = _post_tasks(subtask_name, 'deep_on', natoms)
    subtask_name = os.path.join(iter_name, '02.soft_off')
    e2, err2, tinfo2 = _post_tasks(subtask_name, 'soft_off', natoms)
    fe = fe + e0 + e1 + e2
    print(f'# HTI three-step error [stt_err, sys_err] {err0} {err1} {err2}')
    err = np.sqrt(np.square(err0[0]) + np.square(err1[0]) + np.square(err2[0]))
    sys_err = ((err0[1]) + (err1[1]) + (err2[1]))
    return fe, [err,sys_err], tinfo2


def _print_thermo_info(info) :
    ptr = '# thermodynamics (normalized by nmols)\n'
    ptr += '# E (err)  [eV]:  %20.8f %20.8f\n' % (info['e'], info['e_err'])
    ptr += '# H (err)  [eV]:  %20.8f %20.8f\n' % (info['h'], info['h_err'])
    ptr += '# T (err)   [K]:  %20.8f %20.8f\n' % (info['t'], info['t_err'])
    ptr += '# P (err) [bar]:  %20.8f %20.8f\n' % (info['p'], info['p_err'])
    ptr += '# V (err) [A^3]:  %20.8f %20.8f\n' % (info['v'], info['v_err'])
    ptr += '# PV(err)  [eV]:  %20.8f %20.8f' % (info['pv'], info['pv_err'])
    print(ptr)

def compute_task(job, free_energy_type='helmholtz', scheme='simpson', manual_pv=None, manual_pv_err=None):
    jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))
    fp_conf = open(os.path.join(job, 'conf.lmp'))
    sys_data = lmp.to_system_data(fp_conf.read().split('\n'))
    natoms = sum(sys_data['atom_numbs'])
    jdata = json.load(open(os.path.join(job, 'in.json'), 'r'))
    if 'copies' in jdata :
        natoms *= np.prod(jdata['copies'])
    fe, fe_err, thermo_info = post_tasks(job, natoms)
    _print_thermo_info(thermo_info)

    info = thermo_info.copy()

    pv = None
    pv_err = None
    
    print ('# numb atoms: %d' % natoms)
    print_format = '%20.12f  %10.3e  %10.3e'
    if free_energy_type == 'helmholtz' :
        e1 = fe # e0 + de
        e1_err = fe_err[0]
        print('# Helmholtz free ener per atom (err) [eV]:')
        print(print_format % (fe, fe_err[0], fe_err[1]))
    if free_energy_type == 'gibbs' :
        if manual_pv is None:
            pv = thermo_info['pv']
        else: 
            pv = manual_pv
        if manual_pv_err is None:
            pv_err = thermo_info['pv_err']
        else:
            pv_err = manual_pv_err
        e1 = fe + pv
        e1_err = np.sqrt(fe_err[0]**2 + pv_err**2)
        print('# Gibbs free ener per mol (err) [eV]:')
        print(print_format % (e1, e1_err, fe_err[1]))
    else:
        raise RuntimeError('known free energy type')

    info['free_energy_type'] = free_energy_type
    info['pv'] = pv
    info['pv_err'] = pv_err
    # info['de'] = de
    # info['de_err'] = de_err
    info['e1'] = e1
    info['e1_err'] = e1_err
    with open(os.path.join(job, 'result.json'), 'w') as result:
        result.write(json.dumps(info))
    return info

def _main ():
    parser = argparse.ArgumentParser(
        description="Compute liquid free energy by Hamiltonian TI")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')
    parser_gen.add_argument("-z", "--meam", help="whether use meam instead of dp", action="store_true")

    parser_comp = subparsers.add_parser('compute', help= 'Compute the result of a job')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-t','--type', type=str, default = 'helmholtz', 
                             choices=['helmholtz', 'gibbs'], 
                             help='the type of free energy')
    parser_comp.add_argument('-g', '--pv', type=float, default = None,
                             help='press*vol value override to calculate Gibbs free energy')
    parser_comp.add_argument('-G', '--pv-err', type=float, default = None,
                             help='press*vol error')
    args = parser.parse_args()

    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        output = args.output
        jdata = json.load(open(args.PARAM, 'r'))
        make_tasks(output, jdata, if_meam=args.meam)
    elif args.command == 'compute' :
        compute_task(job=args.JOB, free_energy_type=args.type, manual_pv=args.pv, manual_pv_err=args.pv_err)

     #    fp_conf = open(os.path.join(args.JOB, 'conf.lmp'))
     #    sys_data = lmp.to_system_data(fp_conf.read().split('\n'))
     #    natoms = sum(sys_data['atom_numbs'])
     #    jdata = json.load(open(os.path.join(args.JOB, 'in.json'), 'r'))
     #    if 'copies' in jdata :
     #        natoms *= np.prod(jdata['copies'])
     #    fe, fe_err, thermo_info = post_tasks(args.JOB, natoms)
     #    _print_thermo_info(thermo_info)
     #    print ('# numb atoms: %d' % natoms)
     #    print_format = '%20.12f  %10.3e  %10.3e'
     #    if args.type == 'helmholtz' :
     #        print('# Helmholtz free ener per atom (err) [eV]:')
     #        print(print_format % (fe, fe_err[0], fe_err[1]))
     #    if args.type == 'gibbs' :
     #        if args.pv is None:
     #            pv = thermo_info['pv']
     #        else: 
     #            pv = args.pv
     #        if args.pv_err is None:
     #            pv_err = thermo_info['pv_err']
     #        else:
     #            pv_err = args.pv_err
     #        e1 = fe + pv
     #        e1_err = np.sqrt(fe_err[0]**2 + pv_err**2)
     #        print('# Gibbs free ener per mol (err) [eV]:')
     #        print(print_format % (e1, e1_err, fe_err[1]))

    
if __name__ == '__main__' :
    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from os import path
# from  dpdispatcher import NAME,SHORT_CMD
import setuptools, datetime

today = datetime.date.today().strftime("%b-%d-%Y")
# with open(path.join(NAME, '_date.py'), 'w') as fp :
#     fp.write('date = \'%s\'' % today)

install_requires=['apache-airflow>2.0', 'scipy', 'numpy', 'pymbar', 'dargs', 'dpdispatcher>=0.3']

setuptools.setup(
    name='dpti',
    use_scm_version={'write_to': 'dpti/_version.py'},
    setup_requires=['setuptools_scm'],
    author="Deep Modeling Team",
    author_email="",
    description="Python dpti for thermodynamics integration",
    long_description="",
    long_description_content_type="text/markdown",
    url="",
    python_requires=">3.6",
    packages=['dpti', 'dpti/lib', 'dpti/dags'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
    keywords='free energy thermodynamics integration deepmd-kit',
    install_requires=install_requires,    
    # extras_require={
    #     'docs': ['sphinx', 'recommonmark', 'sphinx_rtd_theme'],
    # },
        entry_points={
        #   'console_scripts': [
        #       SHORT_CMD+'= dpdispatcher.dpdisp:main']
    }
)



import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import soft_param, soft_param_three_element, meam_model
# from potential_common import 
# print(dpti.equi)

class TestEquiForceField(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_deepmd(self):
        input = dict(lamb=0.075, model="graph.pb", sparam=soft_param,
            if_meam=False, meam_model=meam_model)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        """)
        # ret1 = textwrap.dedent(ret1_raw)
        ret2 = dpti.hti._ff_lj_off(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)

    def test_meam(self):
        input = dict(lamb=0.075, model="graph.pb", sparam=soft_param,
            if_meam=True, meam_model=meam_model)

    #     input = dict(lamb=0.075, model=None, sparam=soft_param,
    #         if_meam=True, meam_model=meam_model)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        """)
        ret2 = dpti.hti._ff_lj_off(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)



if __name__ == '__main__':
    unittest.main()
